import click
import http.client


@click.command(
    help="Select based on id or name. Depends on current selection. "
         "If nothing is selected, it will try to select a course, "
         "then an exercise-series, then an exercise. "
         "Will not work with invalid id's or names without flag.")
@click.option("--hidden", "-hidden",
              help="Access a hidden series. Only available when trying to select "
                   "exercise series."
                   "\n\nUsage: dodona select --hidden <TOKEN> <SERIES_ID>")
@click.option("--other", "-other",
              help="Select a course that you're not registered for. Only works with "
                   "an id, not a name.",
              is_flag=True, default=False)
@click.argument('thing')
def select(thing, hidden, other):
    from dodonacli.source import set_data, get_data

    # Read configs in
    config = get_data.get_configs()

    # Start up the connection to Dodona
    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    if config['course_id'] is None:
        config = select_course(connection, headers, thing, config, other)

    elif config['serie_id'] is None:
        if hidden:
            config = select_hidden_series(
                connection, headers, thing, hidden, config
            )
        else:
            config = select_series(connection, headers, thing, config)

    elif config['exercise_id'] is None:
        config = select_exercise(connection, headers, thing, config)

    else:
        # You can't select more when everything is already selected
        print("There is already an exercise selected, please remove selection "
              "with --up or -u to select a new exercise first.")

    # Save selections in config file
    set_data.dump_config(config)

    return


def select_course(connection: http.client.HTTPSConnection, headers: dict,
                  thing: str, config: dict, other: bool):
    if not other:
        return select_registered_course(connection, headers, thing, config)
    else:
        return select_unregistered_course(connection, headers, thing, config)


def select_registered_course(connection: http.client.HTTPSConnection,
                             headers: dict, thing: str, config: dict) -> dict:
    from dodonacli.source import pretty_console, get_data

    # Get all registered courses to check if a valid course was selected
    data_courses = get_data.courses_data(connection, headers)
    courses = {str(course['id']): course['name'] for course in data_courses}

    # First try to match on id, else on name
    if thing.isnumeric() and thing in courses:
        config['course_id'] = thing
        config['course_name'] = courses[thing]

        pretty_console.console.print(
            f"\nCourse [bold]\"{courses[thing]}\"[/] selected.\n")
    else:
        for course in courses.items():
            if thing.lower() in course[1].lower():
                config['course_id'] = course[0]
                config['course_name'] = course[1]

                pretty_console.console.print(
                    f"\nCourse [bold]\"{courses[course[0]]}\" [/] selected.\n")

                # We found it, we can stop searching
                break

    # If config_id is still empty, nothing is found:
    if config['course_id'] is None:
        print("\nNot a valid course-id or -name.\n")

    return config


def select_unregistered_course(connection: http.client.HTTPSConnection,
                               headers: dict, thing: str, config: dict) -> dict:
    import json
    from dodonacli.source import pretty_console, get_data

    # Has to be an id, because it's not possible to match against names 
    if not thing.isnumeric():
        print("The selection needs to be an id (all numbers). "
              "It's not possible to match by name.")
        return config

    link = f"/courses/{thing}"
    connection = get_data.handle_connection_request(
        connection, "GET", link, headers)

    res = connection.getresponse()

    if res.status != 200:
        print("Something went wrong trying to select this course.\n"
              "Make sure you have a valid course-id.\n"
              "\tResonse code: " + str(res.status) +
              "\n\tReason: " + str(res.reason))
        return config

    json_data = json.loads(res.read())
    config['course_id'] = thing
    config['course_name'] = json_data['name']

    pretty_console.console.print(
        f"\nCourse [bold]\"{json_data['name']}\"[/] selected.\n")

    return config


def select_series(connection: http.client.HTTPSConnection, headers: dict,
                  thing: str, config: dict):
    from dodonacli.source import pretty_console, get_data

    data_series = get_data.series_data(connection, headers, config['course_id'])

    series = {str(serie['id']): serie['name'] for serie in data_series}

    if thing.isnumeric() and thing in series:
        config['serie_id'] = thing
        config['serie_name'] = series[thing]
        pretty_console.console.print(
            "\nSeries [bold]\"" + series[thing] + "\"[/] selected.\n")
    else:
        for serie in series.items():
            if thing.lower() in serie[1].lower():
                config['serie_id'] = serie[0]
                config['serie_name'] = serie[1]
                pretty_console.console.print(
                    "\nSerie [bold]\"" + series[serie[0]] + "\"[/] selected.\n")
                break
    if config['serie_id'] is None:
        print("\nNot a valid series id or -name!\n")

    return config


def select_hidden_series(connection: http.client.HTTPSConnection, headers: dict,
                         series_id: str, series_token: str, config: dict):
    import json
    from dodonacli.source import pretty_console, get_data

    if not series_id.isnumeric():
        print("Well this won't work without a valid ID "
              "(only numeric characters). Please try again.")
        return config

    link = f"/series/{series_id}?token={series_token}"
    connection = get_data.handle_connection_request(
        connection, "GET", link, headers)

    res = connection.getresponse()

    if res.status != 200:
        print("Something went wrong trying to select the hidden series.\n"
              "Check if you got the right series-id "
              "(seen in /series/<serie_id> in the link you got).\n"
              "\tResponse code: " + str(res.status) +
              "\n\tReason: " + str(res.reason))
        return config

    json_data = json.loads(res.read())
    config['serie_id'] = series_id
    config['serie_name'] = json_data['name']
    config['serie_token'] = series_token

    pretty_console.console.print(
        f"\nSerie [bold]\"{json_data['name']}\"[/] selected.\n")

    return config


def select_exercise(connection: http.client.HTTPSConnection, headers: dict,
                    selection: str, config: dict) -> dict:
    import textwrap
    from dodonacli.source import pretty_console, get_data

    # Token for hidden series
    if config['serie_token'] is None:
        serie_token = ""
    else:
        serie_token = "?token=" + config['serie_token']

    data_exercises = get_data.exercises_data(
        connection, headers, config['serie_id'], serie_token
    )

    selected_exercise = {}

    if selection.isnumeric() and selection in [exercise['id'] for exercise in data_exercises]:
        selected_exercise = [exercise for exercise in data_exercises if exercise['id'] == selection]

    else:
        for exercise in data_exercises:
            if selection.lower() in exercise['name'].lower():
                selected_exercise = exercise
                break

    if selected_exercise == {}:
        print("\nNot a valid exercise-id or -name.\n")
        return config

    # Store selection
    config['exercise_id'] = str(selected_exercise['id'])
    config['exercise_name'] = selected_exercise['name']
    # Programming language may not exist (f.e. for ContentPage)
    programming_language = selected_exercise.get('programming_language')
    if programming_language:
        config['programming_language'] = programming_language['name']

    pretty_console.console.print(f"\nExercise [bold]{selected_exercise['name']}[/] selected.\n")

    boilerplate = selected_exercise.get('boilerplate')
    if boilerplate and boilerplate.strip() != "":
        print("\nBoilerplate code (copy in boilerplate-file):\n")
        print(textwrap.indent(boilerplate, '\t'))
        with open("boilerplate." + programming_language['extension'], "w") as boilerplate_file:
            boilerplate_file.write(boilerplate)

    return config



import http.client
import json
import markdownify
import re
import shutil
import textwrap

from bs4 import BeautifulSoup
from rich.markdown import Markdown
from rich.padding import Padding

from . import get_data, pretty_console, submission_data_handler


def print_courses_data(json_data: dict, title: str = "Your courses:", prefixes: dict = None):
    """
    Print out the courses in json_data in a neat way
    :param json_data: json object with data about Dodona courses
    :param title: title to display above the courses-list
    :param prefixes: dictionary with a prefix for each id in json_data
    """
    if prefixes is None:
        prefixes = {}

    # List of tuples where each tuple represents a course by id, name and teacher
    display_data = []

    for field in json_data:
        display_data.append((str(field['id']), field['name'], field['teacher']))

    # Find the maximum length of all but the last element in all tuples to align them in the terminal
    max_course_id_length = max(len(e[0]) for e in display_data)
    max_course_name_length = max(len(e[1]) for e in display_data)

    # Print out all courses in display_data
    pretty_console.console.print(f'\n[u bright_blue]{title}[/]')
    for course in display_data:
        pretty_console.console.print(
            (prefixes.get(course[0]) or "\t") +
            f"{course[0].ljust(max_course_id_length)}: "
            f"[bold]{course[1].ljust(max_course_name_length)}[/]\tby {course[2]}"
        )


def print_series_data(json_data: dict, force: bool = False, prefixes: dict = None):
    """
    Print out the exercise-series in json_data in a neat way.
    :param json_data: Json object with data about Dodona exercise-series
    :param force: Boolean to decide if the series description has to be printed, or only a link to it
    :param prefixes: Dictionary with a prefix for each id in json_data
    """
    if prefixes is None:
        prefixes = {}

    # List of tuples where each tuple represents an exercise-series by id, name and description
    display_data = []

    for field in json_data:
        display_data.append(
            (
                str(field['id']),
                field['name'].strip(),
                field['description']
            )
        )

    # Find the maximum length of all but the last element in all tuples to align them in the terminal
    max_series_id_length = max(len(e[0]) for e in display_data)
    max_series_name_length = max(len(e[1]) for e in display_data)

    # Newline for clarity
    print()
    # Print out all the series in display_data while also handling the Markdown inside the series-description
    pretty_console.console.print("[u bright_blue]All series:[/]")
    for series in display_data:
        if force:
            description = series[2].split('\n')
            new_description = ''

            for line in description:
                line = line.rstrip()

                # Remove target pattern from links as they try to open the link in a new tab, not useful for terminal
                line = re.sub(r'{: target="_blank"}', '', line)

                # Replace Markdown bold to Rich Console bold
                line = re.sub(r'\*\*(.*?)\*\*', r'[bold]\1[/bold]', line)

                # Replace Markdown italics to Rich Console italics if it is not in a (link)
                pattern = re.compile(r'([ ,][^ (]*)_(.*?)_([^ ]*[ ,.])')
                line = pattern.sub(r'\1[i]\2[/i]\3', line)

                # Replace Markdown titles to something that appears as a title in terminal
                line = re.sub(r'##+ (.*)', r'[bold white]\1[/bold white]', line)

                # Split lines in multiple when they are too long for the terminal while keeping all lines indented.
                if len(line.replace("[bold]", "").replace("[/bold]", "")) > shutil.get_terminal_size().columns - 8:
                    line = line.split(" ")
                    new_line = ''
                    line_size = 0
                    for word in line:
                        if line_size + len(word) > shutil.get_terminal_size().columns - 8:
                            new_line += '\n'
                            line_size = 0
                        new_line += word + ' '
                        line_size += len(word + ' ')
                    line = new_line

                new_description += line + '\n'

            new_description = textwrap.indent(new_description, '\t')
            pretty_console.console.print(
                f"\t{series[0].ljust(max_series_id_length)}: "
                f"[bold]{series[1].ljust(max_series_name_length)}[/]"
                f"\n{new_description}")
        else:
            pretty_console.console.print(
                (prefixes.get(series[0]) or "\t")
                + f"{series[0].ljust(max_series_id_length)}: "
                + f"[bold]{series[1].ljust(max_series_name_length)}[/]"
            )
    # Newline for clarity
    print()


def print_exercise_data(json_data: dict, prefixes: dict = None):
    """
    Print out the exercises in json_data in a neat way
    :param json_data: json object with data about Dodona exercises in a series
    :param prefixes: dictionary with a prefix for each id in json_data
    """
    if prefixes is None:
        prefixes = {}

    # List of tuples where each tuple represents an exercise by id, name, solved and has_attempt
    display_data = []

    for field in json_data:
        if field['type'] == "ContentPage":
            display_data.append({
                'type': "ContentPage",
                'id': str(field['id']),
                'name': field['name'],
                'has_read': field['has_read']
            })
        elif field['type'] == "Exercise":
            display_data.append({
                'type': "Exercise",
                'id': str(field['id']),
                'name': field['name'],
                # 'last_solution_is_best': field['last_solution_is_best'],
                'has_solution': field['has_solution'],
                # 'has_correct_solution': field['has_correct_solution'],
                'accepted': field['accepted']
            })

    # Find the maximum length of all but the last element in all tuples to align them in the terminal
    max_exercise_id_length = max(len(e['id']) for e in display_data)
    max_exercise_name_length = max(len(e['name']) for e in display_data)

    # Print out all exercises in display_data with indicator about solution-status: solved, wrong or not yet solved
    pretty_console.console.print('\n[u bright_blue]Exercises:[/]')
    for exercise in display_data:
        if exercise['type'] == "Exercise":
            if exercise['accepted']:
                solve_status = "[bold bright_green]SOLVED[/]"
            elif exercise['has_solution']:
                solve_status = "[bold bright_red]WRONG[/]"
            else:
                solve_status = "[bold]NOT YET SOLVED[/]"

            """
            if not exercise['has_solution']:
                solve_status = "[bold]NOT YET SOLVED[/]"
            elif exercise['last_solution_is_best'] and exercise['has_correct_solution']:
                solve_status = "[bold bright_green]SOLVED[/]"
            else:
                solve_status = "[bold bright_red]WRONG[/]"
            """

        elif exercise['type'] == "ContentPage":
            if exercise['has_read']:
                solve_status = "[bold bright_green]READ[/]"
            else:
                solve_status = "[bold]NOT YET READ[/]"

        else:
            solve_status = "[bold]SOLVE STATUS UNKNOWN"

        pretty_console.console.print(
            (prefixes.get(exercise['id']) or "\t")
            + f"{exercise['id'].ljust(max_exercise_id_length)}: "
            + f"[bold]{exercise['name'].ljust(max_exercise_name_length)}[/]\t"
            + solve_status
        )
    print()


def print_exercise(json_data: dict, token: str, force: bool = False):
    """
    Print out the exercise-description.
    Needs to call the Dodona sandbox and convert HTML to text.
    Print out a warning for potential incompleteness, which may be dangerous for tests and exams.
    :param token: API-token as authorization
    :param json_data: json object with info about a Dodona exercise
    :param force: boolean to decide if the exercise description has to be printed, or only a link to it
    """
    if json_data['type'] == 'ContentPage':
        pretty_console.console.print(
            "\nNo need to program anything this time, but you'll have to go read this and mark it as read:\n"
            + json_data['url'].replace(".json", "") + '\n'
        )

    elif not force:
        pretty_console.console.print("\nYou can find the description at \n" + json_data['description_url'] + '\n')

    else:
        # Print the HTML with warnings
        warning = (
                "\n[u bold bright_red]WARNING:[/] the description may be incorrect, "
                "DO NOT rely on this for exams and tests!\n"
                "View in browser: " + json_data['description_url'] + '\n'
        )
        pretty_console.console.print(warning)

        pretty_console.console.print(
            '\n'
            "Expected programming language: " + json_data['programming_language']['name'] +
            '\n'
        )

        # Make sandbox.dodona connection for exercise description:
        sandbox = http.client.HTTPSConnection("sandbox.dodona.be")
        headers = {"Authorization": token}

        stripped_link = json_data['description_url'].replace("https://sandbox.dodona.be", "", 1)
        sandbox.request("GET", stripped_link, headers=headers)

        data = get_data.handle_connection_response(sandbox).decode()

        soup = BeautifulSoup(data, features="html.parser")
        html_description = str(soup.find("div", {"class": "card-supporting-text"}))

        md_description = markdownify.markdownify(html_description)

        md = Markdown(md_description)

        pretty_console.console.print(Padding(md, pad=(0, 0, 0, 3)))

        # Print the HTML with warnings
        pretty_console.console.print(warning)


def print_result(json_results: dict):
    """
    Print out the results of a submission in a neat way
    :param json_results: json object with data about a submission
    """
    if json_results['accepted']:
        # Everything passed, well done!
        pretty_console.console.print("[bold bright_green]All tests passed![/] You can continue to next exercise.")
    else:
        pretty_console.console.print(submission_data_handler.submission_data_handler(json_results))


def print_status(config: dict):
    """
    Print out the current selection of course, exercise-series and exercise.
    :param config: Dictionary with the configs
    """
    pretty_console.console.print(
        f"\n[u bright_blue]Status:[/]\n"
        f"\t{'Course: '.ljust(10)}{config['course_name']}\n"
        f"\t{'Series: '.ljust(10)}{config['serie_name']}\n"
        f"\t{'Exercise: '.ljust(10)}{config['exercise_name']}\n"
    )


def print_exercise_submissions(json_data: dict):
    """
    Print out a list of the (up to) 10 most recent submissions as found in json_data
    :param json_data: Dictionary with submission data
    """
    pretty_console.console.print(
        "\n[u bright_blue]Most recent submissions:[/]"
    )
    for i, submission in enumerate(json_data[:10]):
        if submission['accepted']:
            accepted_emoji = "[bright_green]:heavy_check_mark:[/bright_green]"
        else:
            accepted_emoji = "[bright_red]:heavy_multiplication_x:[/bright_red]"

        status = submission['status']
        if submission['status'] in ("memory limit exceeded", "geheugenlimiet overschreden"):
            status += "\n\t\t\tWow, how did you do that?"

        pretty_console.console.print(
            f"\t{accepted_emoji}  [link={submission['url'].rstrip('.json')}]#{len(json_data) - i: <2}[/link]"
            f"\t{status}\t"
        )

    print()


def print_all_submissions(connection: http.client.HTTPSConnection, headers: dict, json_data: dict):
    """
    Print out a list of the latest 30 submissions for the user, userwide (not tied to an exercise).
    Makes extra requests to Dodona to get the name of the exercises of the submissions
    :param connection: Connection to Dodona
    :param headers: Headers to send with the connection
    :param json_data: Dictionary with submission info
    :return:
    """
    pretty_console.console.print(
        "\n[u bright_blue]Most recent submissions:[/]"
    )

    for i, submission in enumerate(json_data):
        if submission['accepted']:
            accepted_emoji = "[bright_green]:heavy_check_mark:[/bright_green]"
        else:
            accepted_emoji = "[bright_red]:heavy_multiplication_x:[/bright_red]"

        status = submission['status']

        exercise_link = submission['exercise'].replace("https://dodona.be/nl", "", 1)
        connection.request("GET", exercise_link, headers=headers)

        response = connection.getresponse()
        res_status = response.status
        # Need to read always to be able to accept a new request
        result = response.read()

        if res_status == 200:
            exercise_name = json.loads(result).get('name') or "[i]unkown[/i]"
        else:
            exercise_name = "[i]unable to get info about exercise[/i]"

        pretty_console.console.print(
            f"\t{accepted_emoji}  [link={submission['url'].rstrip('.json')}]#{len(json_data) - i: <2}[/link]"
            f"\t{status: <25}"
            f"\t{exercise_name}"
        )

    connection.close()
    print()



import click
from click_default_group import DefaultGroup


@click.group(help="Info about shell-completion, changelog, version, update-availability and GitHub page.",
             cls=DefaultGroup, default='version', default_if_no_args=True)
def info():
    pass


@click.command(help='Display the current version of DodonaCLI. The versioning system '
                    'uses a YYYY.M.D format.')
def version():
    from dodonacli.source import pretty_console
    dodonacli_version = get_dodonacli_version()

    pretty_console.console.print(
        f"DodonaCLI {dodonacli_version}"
    )


@click.command(help='Checks if there is a new update available for DodonaCLI.')
def check_update():
    from packaging.version import parse
    from pkg_info import get_pkg_info
    from dodonacli.source import pretty_console

    dodonacli_version = get_dodonacli_version()

    pkg = get_pkg_info('DodonaCLI')

    if parse(pkg.version) > parse(dodonacli_version):
        pretty_console.console.print(
            f"There is a new version available: {pkg.version}.\n"
            f"You can update your old version ({dodonacli_version}) with"
            "\n\tpip install --upgrade DodonaCLI"
        )
    else:
        print("Your DodonaCLI is up-to-date.")


@click.command(help='Bash completion script for you to source in your .bashrc. '
                    'This is a manual script. '
                    'With the optimesed imports, you can also use the completion provided by the Click library '
                    'for bash, zsh and fish. '
                    'Read more about this at https://click.palletsprojects.com/en/8.1.x/shell-completion/')
def completion():
    print(
        """
# Bash completion script for DodonaCLI, needs to be sourced to work
# Used https://www.gnu.org/software/gnuastro/manual/html_node/Bash-TAB-completion-tutorial.html to help create this

# $1 = name of command -> dodona
# $2 = current word being completed
# $3 = word before word being completed

_dodona(){
    if [ "$3" == "sub" ]; then
        COMPREPLY=( $(compgen -W "load display --help" -- "$2")  )

    elif [ "$3" == "display" ]; then
        COMPREPLY=( $(compgen -W "--force --help" -- "$2") )

    elif [ "$3" == "info" ]; then
        COMPREPLY=( $(compgen -W "version changelog completion github check-update" -- "$2"))

    elif [ "$3" == "select" ]; then
        COMPREPLY=( $(compgen -W "--other --hidden --help" -- "$2") )

    elif [ "$3" == "next" ]; then
        COMPREPLY=( $(compgen -W "--reverse --unsolved --help" -- "$2") )

    elif [ "$3" == "dodona" ]; then
        COMPREPLY=( $(compgen -W "display info next post select status sub tutorial up --help" -- "$2") )

    elif [ "$3" == "post" ]; then
        COMPREPLY=( $(compgen -f -- "$2" | grep -vF ".swp") $(compgen -W "--help --use-link -l -c --check" -- "$2" ))

    elif [[ "$3" =~ ^-[lc]+$ ]] || [ "$3" == "--use-link" ] || [ "$3" == "--check" ]; then
        COMPREPLY=( $(compgen -f -- "$2") )

    elif [ "$3" == "up" ]; then
        COMPREPLY=( $(compgen -W "all top 1 2 3" -- "$2") )

    else
        COMPREPLY=( $(compgen -W "--help") )
    fi
}

complete -F _dodona dodona
        """
    )


@click.command(help='Link to the GitHub page of DodonaCLI. Can be handy for the README page, manpages,'
                    ' Issues (bug reports) and pull requests.')
def github():
    from dodonacli.source import pretty_console

    pretty_console.console.print("https://www.github.com/BWindey/DodonaCLI")


@click.command(help='Changelog for the latest version.')
def changelog():
    from rich.markdown import Markdown
    from dodonacli.source import pretty_console

    changelog_raw = """
- Working on rendering feedback. Please report any troubles you have. 
I may not be able to solve them, but I can definitly hide them =D
- I think this is in its final state, so definitly report all bugs

As always, you can use the "--help" flag after every command and sub-command to learn more.
Happy coding!
    """
    md = Markdown(changelog_raw)
    pretty_console.console.print(md)


def get_dodonacli_version():
    from importlib import metadata
    return metadata.version(__package__.split('.')[0])


info.add_command(version)
info.add_command(check_update)
info.add_command(completion)
info.add_command(github)
info.add_command(changelog)



import http.client
import json
import os
import platform
import socket

from . import set_data, interactive_tutorial


def handle_connection_request(connection: http.client.HTTPSConnection, connection_type: str,
                              link: str, headers: dict) -> http.client.HTTPSConnection:
    """
    Handle of a connection request.
    Quits the program if an exception occurs and displays why.
    :param connection: Connection object
    :param connection_type: "POST", "GET", ...
    :param link: The link to where the connection has to be made
    :param headers: Dictionary to send along the request as headers
    :return: The connection object with the request made
    """
    try:
        connection.request(connection_type, link, headers=headers)

    except socket.gaierror:
        print("Something went wrong trying to connect to Dodona. This is probably an internet connection problem.")
        exit(2)

    return connection


def handle_connection_response(connection: http.client.HTTPSConnection) -> bytes:
    """
    Handle response from Dodona connection
    Terminates program if there was a problem with the connection
    :param connection: HTTPSConnection object that has had a request
    :return: data
    """
    res = connection.getresponse()
    status = res.status
    if status != 200:
        print("Error connecting to Dodona: " + str(status))
        print("Reason: " + res.reason)
        exit(-1)

    data = res.read()
    connection.close()

    return data


def courses_data(connection: http.client.HTTPSConnection, headers: dict):
    """
    Get registered courses of user
    :param connection: HTTPSConnection object to the main Dodona page
    :param headers: Dict with extra info, mainly autorization needed
    :return: json object with info about available courses
    """
    connection = handle_connection_request(connection, "GET", "/courses?tab=my", headers)
    data = handle_connection_response(connection)

    return json.loads(data)


def series_data(connection: http.client.HTTPSConnection, headers: dict, course_id: str):
    """
    Get all exercise-series of course
    :param connection: HTTPSConnection object to the main Dodona page
    :param headers: Dict with extra info, mainly autorization needed
    :param course_id: int id of the course to find series from
    :return: json object with info about available series
    """
    connection = handle_connection_request(connection, "GET", "/courses/" + course_id + "/series", headers=headers)
    data = handle_connection_response(connection)

    return json.loads(data)


def exercises_data(connection: http.client.HTTPSConnection, headers: dict, series_id: str, serie_token: str = ""):
    """
    Get all exercises of exercise-serie
    :param connection: HTTPSConnection object to the main Dodona page
    :param headers: Dict with extra info, mainly autorization needed
    :param series_id: int id of exercise-series
    :param serie_token: Token of a hidden series
    :return: json object with info about available exercises
    """
    connection = handle_connection_request(
        connection,
        "GET",
        "/series/" + series_id + "/activities" + serie_token,
        headers=headers
    )
    data = handle_connection_response(connection)

    return json.loads(data)


def exercise_data(connection: http.client.HTTPSConnection, headers: dict, course_id: str, exercise_id: str):
    """
    Get exercise-info for selected exercise
    :param connection: HTTPSConnection object to the main Dodona page
    :param headers: Dict with extra info, mainly autorization needed
    :param course_id: int id of course
    :param exercise_id: int id of exercise
    :return: json object with info about exercise
    """
    connection = handle_connection_request(
        connection, "GET",
        "/courses/" + course_id + "/activities/" + exercise_id,
        headers=headers
    )
    data = handle_connection_response(connection)

    return json.loads(data)


def exercise_submissions(config: dict, connection: http.client.HTTPSConnection, headers: dict) -> dict:
    """
    Get the last 30 submissions for the current selected exercise
    :param config: Dictionary containing the configs
    :param connection: Connection object to Dodona
    :param headers: Headers dictionary to send along the request
    :return: Dictionary containing the submission data
    """
    course_id = config['course_id']
    series_id = config['serie_id']
    exercise_id = config['exercise_id']

    connection = handle_connection_request(
        connection,
        "GET",
        f"/courses/{course_id}/series/{series_id}/activities/{exercise_id}/submissions",
        headers=headers
    )
    data = handle_connection_response(connection)

    return json.loads(data)


def all_submissions(connection: http.client.HTTPSConnection, headers: dict) -> dict:
    """
    Get the user-wide latest 30 submissions
    :param connection: Connection object to Dodona
    :param headers: Headers dictionary to send along the request
    :return: Dictionary containing the submsision data
    """
    connection = handle_connection_request(
        connection,
        "GET",
        "/submissions",
        headers=headers
    )
    data = handle_connection_response(connection)

    return json.loads(data)


def submission_info(submission_id: int, connection: http.client.HTTPSConnection, headers: dict, config) -> dict:
    """
    Get all the info about the submission together with the name of its exercise.
    :param submission_id: Submission id to get the info about
    :param connection: Connection object to Dodona
    :param headers: Headers dict to send along the request
    :param config: Dictionary with the configs
    :return: Dictionary with the submission data
    """
    connection = handle_connection_request(
        connection,
        "GET",
        f"/submissions/{submission_id}",
        headers=headers
    )
    data = handle_connection_response(connection)
    json_data = json.loads(data)

    exercise_id = json_data['exercise'].split('/')[-1].replace('.json', '')

    exercise = exercise_data(connection, headers, config['course_id'], exercise_id)

    json_data['exercise_name'] = exercise['name']

    return json_data


def get_extension(programming_language: str) -> str:
    """
    Get the file-extension for a programming_language supported by Dodona.
    This is useful when saving code to a file.
    :param programming_language: String representing the name of a programming language
    :return: File-extension as string
    """
    language_dict = {
        "python": "py",
        "sh": "sh",
        "javascript": "js",
        "bash": "sh",
        "java": "java",
        "prolog": "pl",
        "haskell": "hs",
        "text": "txt",
        "csharp": "cs",
        "R": "R",
        "c": "c",
        "Rmarkdown": "Rmd",
        "kotlin": "kt",
        "scheme": "scm",
        "sql": "sql",
        "html": "html",
        "turtle": "py",
        "markdown": "md",
        "ZiM javascript": "txt",
        "cpp": "cpp"
    }
    return language_dict[programming_language]


def get_config_home():
    """
    Returns the path of the config home, this directory stores the config.json file
    :return: The path to the config directory
    """
    # The case switch syntax was introduced in python 3.10
    system = platform.system()
    if system == "Linux":
        platform_config_path = os.getenv("XDG_CONFIG_HOME", default=os.getenv("HOME") + "/.config/")
    elif system == "Darwin":    # aka macOS
        platform_config_path = os.path.join(os.getenv("HOME"), "Library/Application Support")
    elif system == "Windows":
        platform_config_path = os.getenv("APPDATA")
    else:
        return os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../config.json")
    return os.path.join(platform_config_path, "DodonaCLI")


def get_configs():
    """
    Get config data from config.json
    :return: json object with config data
    """
    # Get the path of the config-file.
    config_file_path = os.path.join(get_config_home(), "config.json")

    # fallback to the old path
    needs_migration = False
    if not os.path.exists(config_file_path):
        config_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../config.json")
        needs_migration = True

    # First try to open, if unable to open, create a new config-file and ask user for a token
    try:
        with open(config_file_path, "r") as file:
            config = json.load(file)
            config = validate_config(config)
            if needs_migration:
                set_data.dump_config(config)
                print("config.json has been migrated, the old file is at", os.path.abspath(config_file_path))

    except FileNotFoundError:
        # Create config dictionary
        config = {e: None
                  for e in ["course_id", "course_name", "serie_id", "serie_name", "exercise_id", "exercise_name"]}

        print("\nThis may be your first time using DodonaCLI, do you wish to follow a short tutorial?")
        answer = input("(yes/no): ")

        if answer.lower().startswith("yes"):
            config = interactive_tutorial.start_tutorial(config)
        else:
            TOKEN = input('API-Token not found! Enter your code here: ')
            config["TOKEN"] = TOKEN

        # Save configs
        set_data.dump_config(config)

        exit(0)

    return config


def validate_config(config: dict):
    """
    Checks whether the config file contains all the necessary keys.
    This is needed when an update introduces new keys.
    :param config: Dictionary with the content of config.json
    :return: Updated config dictionary
    """
    keys_to_check = (
        "course_id", "course_name",
        "serie_id", "serie_name",
        "exercise_id", "exercise_name",
        "serie_token", "programming_language"
    )
    for key in keys_to_check:
        if key not in config:
            config[key] = None

    if "TOKEN" not in config:
        print("API token not found.")
        config = get_api_token(config)

    # Not needed any more with the "info update" command
    if "date_update_checked" in config:
        config.pop("date_update_checked")

    return config


def get_api_token(config: dict):
    """
    Asks the user for an API token and saves it
    :param config: Dictionary with the content of config.json
    :return: Updated config dictionary
    """
    config['TOKEN'] = input("Paste your API-token here: ")
    set_data.dump_config(config)

    return config



import click

from dodonacli.commands import select, status, up, post, cli_next
from dodonacli.commands import tutorial, submission, display, info


@click.group(help="A 3rd party Command Line Interface for Dodona. "
                  "Finally you have no need to exit your terminal anymore!\n"
                  "Use --help for more info about flags, or read the README on discord.")
def main():
    """
    A Command Line Interface for Dodona.
    Finally, you have no need to exit your terminal any more!
    Use --help for more info about flags, or read the README on discord.
    """
    pass


main.add_command(display.display)
main.add_command(select.select)
main.add_command(status.status)
main.add_command(up.up)
main.add_command(post.post)
main.add_command(tutorial.tutorial)
main.add_command(submission.sub)
main.add_command(cli_next.cli_next)
main.add_command(info.info)

if __name__ == "__main__":
    # Main entry-point
    main()



"""
This file will contain everything needed to provide a tutorial to first-time users.
It uses "The Coder's Apprentice" featured course on Dodona.
"""
import http.client
import json
import os

import dodonacli.source.get_data
from . import pretty_print, pretty_console, set_data


def start_tutorial(config: dict):
    """
    Runs the tutorial and asks for API token if it wasn't present yet.
    :param config: dict
    :return: config: dict
    """
    # Clear screen for tutorial
    os.system('cls' if os.name == 'nt' else 'clear')
    print(
        """
\n
Welcome to the Dodona CLI, 
this tutorial will show you everything you need to interact 
with Dodona from your terminal.
"""
    )
    config['serie_id'], config['serie_name'] = None, None
    config['exercise_name'], config['exercise_name'] = None, None

    if "TOKEN" not in config:
        pretty_console.console.print(
            """
\n\n
Before we begin, we'll need an API-token to authorize your 
connection with Dodona. 
You can find it on your profile page on https://dodona.be
"""
        )
        config = dodonacli.source.get_data.get_api_token(config)

    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    config = tutorial_select_course(config, connection, headers)
    config = tutorial_select_series(config, connection, headers)
    config = tutorial_select_exercise(config, connection, headers)

    tutorial_view_exercise(config, connection, headers)
    tutorial_post_exercise(config, connection, headers)

    config = tutorial_conclude(config)

    return config


def tutorial_handle_connection(config: dict, connection: http.client.HTTPSConnection):
    res = connection.getresponse()
    status = res.status

    if status != 200:
        if status == 401:
            print("Authorization didn't work, "
                  "please re-enter your API-token and make sure it's correct.")
            config = dodonacli.source.get_data.get_api_token(config)
            set_data.dump_config(config)

        else:
            pretty_console.console.print(
                "Error connecting to Dodona: " + str(status)
            )
            pretty_console.console.print("Reason: " + res.reason)

        pretty_console.console.print(
            "Tutorial will terminate due to error, "
            "you can start again with `dodona tutorial`")
        exit(-1)

    data = res.read()
    connection.close()

    return json.loads(data)


def tutorial_select_course(config: dict, connection: http.client.HTTPSConnection,
                           headers: dict):
    pretty_console.console.print(
        "\n\n\n"
        "Use the command `dodona display` to show the available courses."
    )

    command = input("$ ")

    while command.rstrip() != "dodona display":
        print("That was not the right command, please try again")
        command = input("$ ")

    connection.request("GET", "/courses?tab=featured", headers=headers)
    json_data = tutorial_handle_connection(config, connection)

    pretty_print.print_courses_data(json_data, "Featured courses")

    pretty_console.console.print(
        """
Select now "The Coder's Apprenctice" with `dodona select` + 
the courses id, or (distinct part of) the courses name
"""
    )
    command = input("$ ")

    while not command.rstrip().startswith("dodona select"):
        print("That was not the right command, please try again")
        command = input("$ ")

    if (len(command.split()) < 3
            or command.split()[2] != "296"
            and command.split()[2].lower() not in "The Coder's Apprentice".lower()):
        pretty_console.console.print(
            """
Watch out, you used a wrong id or name to select 
"The Coder's Apprentice"!\n
The tutorial will continue as if you selected it right, 
but pay attention next time.
"""
        )

    config['course_id'] = 296
    config['course_name'] = "The Coder's Apprentice"

    pretty_console.console.print(
        """
\n
The course is now selected, 
use `dodona status` to view your selection:
"""
    )
    command = input("$ ")

    while not command.rstrip() == "dodona status":
        print("That was not the right command, please try again")
        command = input("$ ")

    pretty_print.print_status(config)
    print("Fantastic, let's move on to selecting an exercise series.")

    return config


def tutorial_select_series(config: dict, connection: http.client.HTTPSConnection,
                           headers: dict):
    pretty_console.console.print(
        """
\n\n\n
Use the command `dodona display` to show the available 
exercise-series.
"""
    )
    command = input("$ ")

    while command.rstrip() != "dodona display":
        print("That was not the right command, please try again")
        command = input("$ ")

    connection.request(
        "GET", "/courses/" + str(config['course_id']) + '/series',
        headers=headers)
    json_data = tutorial_handle_connection(config, connection)

    # To prevent the screen being blasted with a lot of text, 
    # only print out the first 6 exercise series
    pretty_print.print_series_data(json_data[:6])

    pretty_console.console.print(
        """
Select now "2. Using Python" with `dodona select` + 
the series' id, or (distinct part of) the series' name.
"""
    )
    command = input("$ ")

    while not command.rstrip().startswith("dodona select"):
        print("That was not the right command, please try again")
        command = input("$ ")

    if (len(command.split()) < 3
            or command.split()[2] != "2592"
            and command.split()[2].lower() not in "2. Using Python".lower()):
        pretty_console.console.print(
            """
Watch out, you used a wrong id or name to select 
"2. Using Python"!\n
The tutorial will continue as if you selected it right, 
but pay attention next time.
"""
        )

    config['serie_id'] = 2592
    config['serie_name'] = "2. Using Python"

    pretty_console.console.print(
        """
\n
The exercise-series is now selected, use `dodona status` 
to view your selection:
"""
    )
    command = input("$ ")

    while not command.rstrip() == "dodona status":
        print("That was not the right command, please try again")
        command = input("$ ")

    pretty_print.print_status(config)
    print("Fantastic, let's move on to selecting an exercise.")

    return config


def tutorial_select_exercise(config: dict, connection: http.client.HTTPSConnection,
                             headers: dict):
    pretty_console.console.print(
        """
\n\n\n
Use the command `dodona display` to show the available exercises.
"""
    )
    command = input("$ ")

    while command.rstrip() != "dodona display":
        print("That was not the right command, please try again")
        command = input("$ ")

    connection.request(
        "GET", "/series/" + str(config['serie_id']) + '/activities',
        headers=headers)
    json_data = tutorial_handle_connection(config, connection)

    pretty_print.print_exercise_data(json_data)

    pretty_console.console.print(
        """
Select now "Hello, World!" with `dodona select` + 
the series' id, or (distinct part of) the series' name.
"""
    )
    command = input("$ ")

    while not command.rstrip().startswith("dodona select"):
        print("That was not the right command, please try again")
        command = input("$ ")

    if (len(command.split()) < 3
            or command.split()[2] != "1399231809"
            and command.split()[2].lower() not in "Hello, World!".lower()):
        pretty_console.console.print(
            """
Watch out, you used a wrong id or name to select 
"2. Using Python"!\n
The tutorial will continue as if you selected it right, 
but pay attention next time.
"""
        )

    config['exercise_id'] = 1399231809
    config['exercise_name'] = "Hello, World!"

    # This exercise has boilerplate code
    with open('boilerplate.py', 'w') as boilerplate:
        boilerplate.write(json_data[5]['boilerplate'])
    print(
        """
This exercise has some boilerplate code attached to it, 
you can view it in the boilerplate file, or here:\n|\t        
""" + json_data[5]['boilerplate']
    )

    pretty_console.console.print(
        """
\n
The exercise is now selected, 
use `dodona status` to view your selection:
"""
    )
    command = input("$ ")

    while not command.rstrip() == "dodona status":
        print("That was not the right command, please try again.")
        command = input("$ ")

    pretty_print.print_status(config)
    print("Fantastic, let's move on to viewing an exercise description.")

    return config


def tutorial_view_exercise(config: dict, connection: http.client.HTTPSConnection,
                           headers: dict):
    pretty_console.console.print(
        """
\n\n\n
Use the command `dodona display` to show the description of the 
exercise.
"""
    )
    command = input("$ ")

    while command.rstrip() != "dodona display":
        print("That was not the right command, please try again")
        command = input("$ ")

    connection.request("GET", "/courses/" + str(config['course_id'])
                       + "/series/" + str(config['serie_id'])
                       + '/activities/' + str(config['exercise_id']),
                       headers=headers)
    json_data = tutorial_handle_connection(config, connection)

    pretty_print.print_exercise(json_data, config['TOKEN'])


def tutorial_post_exercise(config: dict, connection: http.client.HTTPSConnection,
                           headers: dict):
    pretty_console.console.print(
        """
\n\n\n
Now you can post the solution. You don't need to write any code 
for it, as it is already writtin in the 'boilerplate'-file. 
You can post it by writing `dodona post <SOLUTION_FILE_NAME>`.
Replace <SOLUTION_FILE_NAME> with the correct file-name, 
in this case 'boilerplate'.
"""
    )
    command = input("$ ")

    while command.rstrip() != "dodona post boilerplate":
        print("That was not the right command, please try again")
        command = input("$ ")

    with open('boilerplate', 'r') as solution_file:
        set_data.post_solution(solution_file.read(), connection, headers, config['course_id'], config['exercise_id'])


def tutorial_conclude(config: dict) -> dict:
    pretty_console.console.print(
        """
Almost done, now deselect everything. You can do this with the 
'up' command + a selection from [1|2|3|all|top]. The numbers specify 
the amount of levels you'll deselect, while all/top deselects 
everything at once.
"""
    )

    command = input("$ ")

    # Parse input
    while not command.rstrip().startswith("dodona up"):
        print("That wasn't the right command, please try again.")
        command = input("$ ")

    # Deselect everything (except the token)
    all_configs = list(config.keys())
    all_configs.remove("TOKEN")

    for conf in all_configs:
        config[conf] = None

    # Print the right message
    amount = command.replace("dodona up", "").strip()
    if amount in ("all", "top", "3"):
        print("Deselected everything")
    elif amount.isnumeric():
        print("Deselected everything to save on time. Normally this would "
              "deselect " + amount + "levels")
    else:
        print("You didn't enter a valid amount, but I'll let it slip now.\n"
              "Deselected everything.")

    # End of tutorial
    pretty_console.console.print(
        """
Remember you can always use `dodona [subcommand] --help` to get 
help with all the subcommands. Additionally there is a manual-page 
you can download from [link=https://www.github.com/BWindey/DodonaCLI]
GitHub[/link]. You can also find a completion script there.
Installation instructions for both can be found in the README.md.
"""
    )

    return config



import click


@click.command(help="Deselect default last selected thing. " 
                    "Can be used with an argument to deselect everything with "
                    "'all' or 'top', or deselect 1, 2 or 3 levels.")
@click.argument('amount', default='1',
                type=click.Choice(['all', 'top', '1', '2', '3'], 
                                  case_sensitive=False))
def up(amount):
    from dodonacli.source import set_data, get_data
    # Read configs in
    config = get_data.get_configs()

    if str(amount).lower().strip() in ('all', 'top'):
        for e in ('exercise_id', 'exercise_name', 'serie_id', 'serie_name', 'serie_token', 'course_id', 'course_name'):
            config[e] = None
        print("\nDeselected everything.\n")

    elif not isinstance(amount, int) and not amount.isnumeric():
        print("You didn't provide a number or 'all', nothing deselected.")

    else:
        levels = int(amount)
        print()
        for _ in range(levels):
            if config.get('exercise_id') is not None:
                config['exercise_id'] = None
                config['exercise_name'] = None
                config['programming_language'] = None
                print("Deselected exercise.")
            elif config.get('serie_id') is not None:
                config['serie_id'] = None
                config['serie_name'] = None
                config['serie_token'] = None
                print("Deselected series.")
            elif config.get('course_id') is not None:
                config['course_id'] = None
                config['course_name'] = None
                print("Deselected course")
            else:
                print("Nothing selected.")
                break
        print()

    set_data.dump_config(config)
